/* $Id: Ifft2.java,v 1.39 2017/03/27 23:27:19 mit3 Exp $ */
/* Ifft2.java v0.3.1*/
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.*;
import java.text.DecimalFormat;

public class Ifft2 extends Applet 
    implements ItemListener, ActionListener, AdjustmentListener,
	       MouseListener, MouseMotionListener
{
    final int Oct = 12;
    final int KB = 88;
    final int KA = KB+Oct*3;
    final int Cent = 1200;
    final int MAXPARTIAL = 20;

    final int FFTX = 0;
    final int WAVE = 1;
    final int BEAT = 2;
    final int CENT = 3;
    final int INHA = 4;

    final double MC = Cent/Math.log(2);
    final double Log10 = Math.log(10);
    final Color Cone = Color.cyan;      // select.key
    final Color Ctwo = Color.red;       // fund.key
    final Color Cnot = Color.lightGray; // off.key
    private static final long serialVersionUID = 1L;

    int W, H;
    int H2, HS;
    int FNT;
    int N0 = 11; // 2^n(12-4096 13-8192 14-16384 15-32768)
    int N = (int)Math.pow(2, N0); // Sampling 
    int N2 = N/2;
    int Wbyf = 2;      // wide fft
    int Wbyw = 1;      // wide wave
    int Wbyb = 32;     // wide beats
    int PA = 1;        // partial# 1~20
    int GRAP = FFTX;
    int Smplf = 11025; // 44100, 22050, 11025, 8000
    int AA = 440;
    int Cmax = 100; //cent
    int Bmax = 10;  // beat
    int Dot;
    int Dsf;
    int KEY1;
    int KEY2;
    double DK;          // w/keys
    double Gainf = 0.2; // gain fft
    double Gainw = 30;  // gain wave
    double Time = 2.0;  // wave data
    double CentTemp;    // centݻ
    double Cby = 1.0;   // centܿ
    boolean DSsw;       // false:ch1 true:ch2
    boolean CH1, CH2;
    boolean GTsw;       // graph./text
    boolean MOUSEsw;    // mouse sw.
    boolean TUNEsw;
    boolean ABSsw;
    boolean SPsw;

    double[] Re;    // for fft
    double[] Im;    // "
    double[] Dat;   // "
    double[] Freq1;
    double[] Freq2;
    double[] Beat;
    DecimalFormat df1, df2;

    final String[] sp_list = 
    {"14","13","12","11","10"}; // 16384 8192 4096 2048 1024
    final String[] sf_list = 
    {"11025","22050","44100","8000"}; // sampl
    final String[] Boxes = {"FFT","Wave","Beats","Cent","Inha."};
    final String[] Octnm = {"A2","A1","A","a","a1","a2","a3","a4"};
    final String[] Keynm =
    {"A..","B..","H..","C..","Cis","D..","Es.","E..","F..","Fis","G..","Gis"};

    private Choice chP;        // partial
    private Choice chS;        // sampling N
    private Choice chF;        // sampling F
    private Button btCm, btCp; // /, X (cent)
    private Button btCH, btON; // key1/2, on/off
    private Button btU, btD;   // +,- level
    private Button btW, btN;   // <,> wide
    private Button btGT;       // grap./text
    private Button btTune;
    private Button btAbs;
    private Button btSP;
    private Button[] btWest;
    private TextField tfA;
    private TextArea taA;
    private Scrollbar sbCent;
    private Canvas can;
    private CardLayout Card;
    private Panel deck;
    private Panel NP, EP, SP, WP;
    private Graphics gi;
    private Image ig;
    private Font Fnt;

    MakeInharmonicity MI;
    TuningTriads TT;
    MakeWaves2 MW;
    Interval IN;
    FFT fftN;

    public void init() {
	String fnt = getParameter("font");
	FNT = (fnt != null) ? Integer.parseInt(fnt) : 12;
	Fnt = new Font("SansSerif", Font.PLAIN, FNT);
	setFont(Fnt);
	df1 = new DecimalFormat("#0.0###");
	df2 = new DecimalFormat("#0.0##");

	buildLayout();
	init_start();
    }
    public void buildLayout() {
	setBackground(Color.lightGray);
	setLayout(new BorderLayout());

	deck = new Panel();
        Card = new CardLayout();
        deck.setLayout(Card);

	NP = new Panel();
	NP.setLayout(new GridLayout(2,1,0,0));
	add(NP, "North");

        Panel u1 = new Panel();
        u1.setLayout(new FlowLayout());
        NP.add(u1);

        Panel u2 = new Panel();
        u2.setLayout(new FlowLayout());
        NP.add(u2);

	tfA = new TextField(45);
	tfA.setFont(Fnt);
        tfA.setEditable(false);
        tfA.addActionListener(this);
        u1.add(tfA);
	//

	chP = new Choice();
	for (int i = 0; i < MAXPARTIAL; i++)  // 1~20
	    chP.addItem(""+(i+1));
	chP.addItemListener(this);
	u2.add(chP);

	btTune = new Button("Tune3");
	btTune.addActionListener(this);
	u2.add(btTune);

	btAbs = new Button("abs.");
	btAbs.addActionListener(this);
	u2.add(btAbs);

	btSP = new Button("s.p.");
	btSP.addActionListener(this);
	u2.add(btSP);

	chS = new Choice();
	for (int i = 0; i < sp_list.length; i++)  // 14~
	    chS.addItem(sp_list[i]);
	chS.addItemListener(this);
	N0 = 11;
	u2.add(chS);

	chF = new Choice();
	for (int i = 0; i < sf_list.length; i++)  // 11025~
	    chF.addItem(sf_list[i]);
	chF.addItemListener(this);
	u2.add(chF);
	//-----------------------------u1/u2

	WP = new Panel();
	WP.setLayout(new GridLayout(Boxes.length+2,1,0,0));
        add(WP, "West");

	btWest = new Button[Boxes.length];
	for (int i = 0; i < btWest.length; i++) {
	    btWest[i] = new Button(Boxes[i]);
            btWest[i].addActionListener(this);
            WP.add(btWest[i]);
        }

	WP.add(new Label());

 	btGT = new Button("Grap.");
	btGT.addActionListener(this);
	WP.add(btGT);
	//---

	EP = new Panel();
	EP.setLayout(new GridLayout(1,2,0,0));
	add(EP, "East");

	Panel r0 = new Panel();
	r0.setLayout(new GridLayout(1,1,0,0));
	EP.add(r0);

	Panel r1 = new Panel();
	r1.setLayout(new GridLayout(5,1,0,0));
	EP.add(r1);

	sbCent = new Scrollbar(Scrollbar.VERTICAL,50,10,0,110);
        r0.add(sbCent);
        sbCent.addAdjustmentListener(this);
	//

	btU = new Button("+");
	btU.addActionListener(this);
	r1.add(btU);

	btD = new Button("-");
	btD.addActionListener(this);
	r1.add(btD);

	r1.add(new Label());

	btW = new Button(">");  // wide
	btW.addActionListener(this);
	r1.add(btW);

	btN = new Button("<");  // narrow
	btN.addActionListener(this);
	r1.add(btN);
	//--------------------------right/down

	SP = new Panel();
	SP.setLayout(new GridLayout(1,5,0,0));
	add(SP, "South");

	btCH = new Button("Key1");
	btCH.addActionListener(this);
	SP.add(btCH);

	btON = new Button("ON ");
	btON.addActionListener(this);
	SP.add(btON);

	SP.add(new Label());

	btCp = new Button("/");
	btCp.addActionListener(this);
	SP.add(btCp);

	btCm = new Button("X");
	btCm.addActionListener(this);
	SP.add(btCm);
	//------------------------down/center

	can = new myCanvas();
        can.addMouseListener(this);
        can.addMouseMotionListener(this);

	Panel card1 = new Panel();
	card1.add(can);
	deck.add(card1, "card1");

        taA = new TextArea("", 16, 50, TextArea.SCROLLBARS_BOTH);
	taA.setFont(Fnt);
        taA.setEditable(false);

        ScrollPane card2 = new ScrollPane();
	card2.add(taA);
	deck.add(card2, "card2");

        add(deck, "Center");
    }
    public void start() {
	W = getCanvasD().width;
	H = getCanvasD().height;

	can.setSize(W, H);
	ig = createImage(W, H);
        gi = ig.getGraphics();
        gi.setFont(Fnt);

	H2 = H/2;
	HS = H-FNT*3;
	DK = (double)W/KB;
	set_dot();
    }
    public Dimension getCanvasD() {
        Dimension du = NP.getSize();
	Dimension de = EP.getSize();
	Dimension ds = SP.getSize();
	Dimension dw = WP.getSize();
	Dimension da = getSize();
        int cw = da.width-(de.width+dw.width);
        int ch = da.height-(du.height+ds.height)-6;
        Dimension can = new Dimension(cw, ch);

        return can;
    }
    public void init_start() {
	IN = new Interval();
	MI = new MakeInharmonicity();
	TT = new TuningTriads();
	TT.setInha(MI.getInha());
	set_nn();
	MW = new MakeWaves2(Smplf, Time, PA);

	chS.select(sp_list[3]); // 11
	KEY1 = 37-1;
	KEY2 = KEY1-Oct;

	init_beat();
	reset_tune();
	calc_wave();
	set_enabled();
    }
    public void set_nn() {
        N = (int)Math.pow(2, N0);
	N2 = N/2;
        fftN = new FFT(N);

	Re = new double[N];
        Im = new double[N];
        Dat = new double[N2];
    }
    public void reset_tune() {
        set_tune();
	set_scrCent();

	MW.makeWaveOff();
    }
    public void set_tune() {
	TUNEsw = !TUNEsw;

        if (TUNEsw) {
            TT.Tuning();
        } else {
            TT.initCent();
	}

        btTune.setLabel(TUNEsw ? "Clear" : "Tune3");
    }
    //
    public void mouseMoved(MouseEvent me) {
	if (MOUSEsw && (GRAP == CENT || GRAP == INHA)) {
	    int key = (int)(me.getX()/DK);
	    if (DSsw)
		KEY2 = key;
	    else
		KEY1 = key;

	    set_scrCent();
	    MW.setKey(KEY1, KEY2);
	    can.repaint();
        }
    }
    public void set_scrCent() {
        sbCent.setValue(50); // reset

        CentTemp = TT.getCent(get_key());
    }
    public void mouseDragged(MouseEvent me) {}
    public void mousePressed(MouseEvent m) {
	MOUSEsw = !MOUSEsw;
	can.repaint();
    }
    public void mouseClicked(MouseEvent m) {}
    public void mouseReleased(MouseEvent m){}
    public void mouseEntered(MouseEvent m) {}
    public void mouseExited (MouseEvent m) {}
    public void adjustmentValueChanged(AdjustmentEvent e) {
	if (e.getSource() == sbCent)
	    set_cent();

	disp_sw();
    }
    public void itemStateChanged(ItemEvent e) {
	if (e.getSource() == chP) {             // partial#
	    String pa = chP.getSelectedItem();
	    PA = Integer.parseInt(pa);

	    MW.setMulti(PA);
	} 
       
	if (e.getSource() == chS) {              // N 
	    String sp = chS.getSelectedItem();
	    N0 = Integer.parseInt(sp);
	    set_nn();
	}

	if (e.getSource() == chF) {              // sampling
	    String sf = chF.getSelectedItem();
	    Smplf = Integer.parseInt(sf);

	    MW.setSplgf(Smplf);
	}

	set_enabled();
	disp_sw();
    }
    public void actionPerformed(ActionEvent e) {  
	if (e.getSource() == btCH)        // key1/2
	    set_ch();
	
	else if (e.getSource() == btON)   // on/off
	    set_onoff();

	else if (e.getSource() == btCm)   // cent/
	    pm_cent(false);
	else if (e.getSource() == btCp)   // centX
	    pm_cent(true);

	else if (e.getSource() == btU)    // gain up
	    set_gain(true);
	else if (e.getSource() == btD)    // gain down
	    set_gain(false);

	else if (e.getSource() == btW)    // wide
	    set_wide(true);
	else if (e.getSource() == btN)    // narrow
	    set_wide(false);

	else if (e.getSource() == btGT)   // graph./text
	    set_graph();

	else if (e.getSource() == btTune)
	    reset_tune();

	else if (e.getSource() == btAbs)
	    set_abs();

	else if (e.getSource() == btSP)
	    set_sp();

	for (int i = 0; i < Boxes.length; i++) { // disp sw.
	    if (e.getSource() == btWest[i]) {
		GRAP = i;
		break;
	    }
	}

	set_enabled();
	disp_sw();
    }
    public void set_sp() {
	SPsw = !SPsw;
	btSP.setLabel(SPsw ? "S.P." : "s.p.");

	MW.setSP(SPsw);
    }
    public void set_abs() {
	ABSsw = !ABSsw;

	btAbs.setLabel(ABSsw ? "ABS." : "abs.");
    }
    public void set_cent() {
        double ce = (50-sbCent.getValue())*Cby;
        TT.setCent(get_key(), ce+CentTemp);

	MW.makeWaveOff();
    }
    public void pm_cent(boolean sw) {
	Cby = (sw) ? Cby/2.0 : Cby*2.0;

	set_scrCent();
	double cb = 50*Cby;
	showStatus("Scrollbar: -"+dform(cb)+":"+dform(cb)+"[cent]");
    }
    public void disp_sw() {
	calc_wave();

	if (GTsw)
	    re_write();
	else
	    can.repaint();
    }
    public void calc_wave() {
	if (!MW.isMakeWave())
	    mwave();
    }
    public void calc_fft() {
	if (MW.isMakeWave()) {
	    gauss();
	    fft();
	}
    }
    public void set_enabled() {
	boolean is = !GTsw && !(GRAP == INHA);
	btU.setEnabled(is); // +
	btD.setEnabled(is); // -

	boolean cs = !GTsw && is && !(GRAP == CENT);
	btW.setEnabled(cs); // <
	btN.setEnabled(cs); // >

	boolean as = (GRAP == BEAT);
	btAbs.setEnabled(as);
    }
    public void set_onoff() {
	if (DSsw)
	    CH2 = !CH2;
	else
	    CH1 = !CH1;

	set_on_button();
	MW.setCH(CH1, CH2);
    }
    public void set_on_button() {
	btON.setLabel(get_on() ? "OFF" : "ON ");
    }
    public boolean get_on() {
	return DSsw ? CH2 : CH1; 
    }
    public String get_ch(boolean sw) {
	return sw ? "Key2" : "Key1";
    }
    public int get_key() {
	return DSsw ? KEY2 : KEY1;
    }
    public void set_graph() {
	GTsw = !GTsw;
	btGT.setLabel(GTsw ? "Text" : "Graph");
	Card.next(deck);
    }
    public void set_ch() {
	DSsw = !DSsw;
	btCH.setLabel(get_ch(DSsw));

	set_scrCent();
	set_on_button();
    }
    public void set_wide(boolean sw) {
	switch (GRAP) {
	case WAVE:
	    Wbyw = limit(1, 1024, sw ? Wbyw/2 : Wbyw*2);
	    break;
	case FFTX:
	    Wbyf = limit(2,  512, sw ? Wbyf/2 : Wbyf*2);	
	    break;
	case BEAT:
	    Wbyb = limit(2,16384, sw ? Wbyb*2 : Wbyb/2);
	    break;
	default:
	}
    }
    public int limit(int min, int max, int val) {
	if (val > max)
	    val = max;
	else if (val < min)
	    val = min;
	return (val);
    }
    public void set_gain(boolean sw) {
	double ms2 = Math.sqrt(2);

	switch (GRAP) {
	case WAVE:
	    Gainw = sw ? Gainw*ms2 : Gainw/ms2;
	    break;
	case FFTX:
	    Gainf = sw ? Gainf*ms2 : Gainf/ms2;
	    break;
	case BEAT:
	    Bmax = sw ? Bmax*2 : Bmax/2;
            int min = 1;
            if (Bmax < min)
                Bmax = min;
	    break;
	case CENT:
	    Cmax = limit(10, 400, sw ? Cmax*2 : Cmax/2);
	    break;
	default:
	}
    }
    public void mwave() {
	if (MW.isMakeWave())
            return;

	Freq1 = make_freq(KEY1);
	Freq2 = make_freq(KEY2);
	MW.calcWave(Freq1, Freq2);

	set_beats();
    }
    public double[] make_freq(int key) {
	double[] freq = new double[PA+1];

	for (int i = 1; i <= PA; i++) {
	    freq[i] = getf(key, i);
	}

	return freq;
    }
    public void init_beat() {
	Beat = new double[PA+1];
    }
    public void set_beats() {
	double be, min;
	init_beat();

	for (int i = 1; i <= PA; i++) {
	    min = Integer.MAX_VALUE;

	    for (int j = 1; j <= PA; j++) {
		be = Freq1[i]-Freq2[j];
		if (Math.abs(min) > Math.abs(be))
		    min = be;
	    }

	    Beat[i] = min;
	}
    }
    //-----------------------------

    class myCanvas extends Canvas {
	private static final long serialVersionUID = 1L;
	public void update (Graphics g) {
	    can.paint(g);
	}
	public void paint(Graphics g) {
            paintI();
            g.drawImage(ig, 0, 0, this);
        }
    }
    public void paintI() {
	blackw(gi);

	switch (GRAP) {
	case WAVE:
	    dispWave(gi);
	    break;
	case BEAT:
	    dispBeats(gi);
	    break;
	case FFTX:
	    dispFft(gi);
	    break;
	case CENT:
	    dispCent(gi);
	    break;
	case INHA:
	    dispInha(gi);
	    break;
	default:
	}

	text_mouse();
    }
    public void blackw(Graphics g) {
	g.setColor(Color.black);
	g.fillRect(0, 0, W, H);
    }
    public void text_mouse() {
	StringBuffer sb = new StringBuffer();
	sb.append(getkey(DSsw));

	String ints = IN.getIinter(KEY2-KEY1);
	sb.append((ints.equals("") ? "" : "("+ints+")"));

	tfA.setText(sb.toString());
    }
    public void set_dot() {
	Dot = (int)DK;
	Dsf = Dot/2;
    }
    //
    public void dispInha(Graphics g) {
	Color co = Color.white;
	int hb = H/5;

	disp_log(g, hb);
	disp_aline(g);

	for (int i = 0; i < KB; i++) {
	    g.setColor(change_color(i, co));
	    g.fillOval((int)(i*DK-Dsf), 
		       (int)(H2-log_base(geti(i))*hb-Dsf), Dot, Dot);
	}
    }
    public void disp_log(Graphics g, int hb) {
	double h0 = 0.001, hl;
	int hm;

	for (int i = 1; i < 6; i++) { // ʿ
	    hl = h0*Math.pow(10, i);
	    hm = (int)(H2-hb*log_base(hl));

	    g.setColor(Color.blue);
	    g.drawLine(0, hm, W, hm);

	    g.setColor(Color.orange); // 
	    g.drawString(""+(float)hl, 2, hm);
	}
    }
    //
    public void dispCent(Graphics g) {
	double dx = (double)W/KB;
	Color co = Color.white;
	double dh = (double)H/Cmax;

	disp_scale(g, dh);
	disp_aline(g);
	disp_center_line(g);

	for (int i = 0; i < KB; i++) {
	    g.setColor(change_color(i, co));
	    g.fillOval((int)(dx*i-Dsf),
		       (int)(H2-getc(i)*dh-Dsf), Dot, Dot);
	}
    }
    public Color change_color(int key, Color co) {
	Color re = co;

	if (key == KEY1)
	    re = set_color(false);
	else if (key == KEY2)
	    re = set_color(true);

	return re;
    }
    public Color set_color(boolean sw) {
	if (sw)
	    return CH2 ? Cnot : Cone;
	else
	    return CH1 ? Cnot : Ctwo;
    }
    public void disp_scale(Graphics g, double dh) {
        int hx = Cmax/10, xs = 2, hp, hm, hi;
        double hy;

        for (int i = 1; i < 6; i++) { // scale line
            hi = hx*i;
            hy = dh*hi;
            hp = (int)(H2+hy);
            hm = (int)(H2-hy);

            g.setColor(Color.blue);
            g.drawLine(0, hm, W, hm);
            g.drawLine(0, hp, W, hp);

            g.setColor(Color.yellow);
            g.drawString(String.valueOf(hi), xs+2, hm+FNT);
            g.drawString("-"+hi, xs, hp);
        }
    }
    public void disp_center_line(Graphics g) {
        g.setColor(Color.gray);
        g.drawLine(0, H2, W, H2);
    }
    public void disp_aline(Graphics g) {
	int wi;

	for (int i = 0; i < Octnm.length; i++) {
	    wi = (int)(DK*i*Oct);

	    g.setColor(Color.blue);
	    g.drawLine(wi, 0, wi, H);

	    g.setColor(Color.red);
	    g.drawString(Octnm[i], wi, H-FNT);
	}
    }
    //
    public void dispWave(Graphics g) {
	int x1, y1;
	int alldata = (int)(Time*Smplf);
	int Nmax = alldata/Wbyw;
	double dn = W/(double)Nmax;

	disp_center_line(g);

	g.setColor(Color.green);
	for (int i = 0; i < Nmax; i++) {
	    x1 = (int)(dn*i);
	    y1 = (int)(H2-MW.getWave(i)*Gainw);
	    g.drawLine(x1, y1, x1, y1);
	}

	disp_wave_wide(g, Nmax, alldata);
    }
    public void disp_wave_wide(Graphics g, int maxw, int all) {
        StringBuffer sb = new StringBuffer(64);

        sb.append(dform(100.0*maxw/all));
        sb.append("% "+dform(maxw/(double)Smplf));
        sb.append("[sec]");

        g.setColor(Color.lightGray);
        g.drawString(sb.toString(), 1, FNT);
    }
    //
    public void dispBeats(Graphics g) {
	int dot = (int)(W/MAXPARTIAL/2.0);
	int dsf = dot/2;
	int h2 = ABSsw ? H : H2;
	double be, dp = W/(double)PA;
	double db = h2/(double)Bmax;

	disp_beat_vline(g, dp, dsf);
	disp_beat_line(g);

	g.setColor(Color.cyan);
	for (int i = 0; i < PA; i++) {
	    be = ABSsw ? Math.abs(Beat[i+1]) : Beat[i+1];

	    g.fillOval((int)(i*dp), (int)(h2-be*db-dsf), dot, dot);
	}
    }
    public void disp_beat_line(Graphics g) {
	int hm, hp = 0, xs = 2, max = 5;
	int h2 = ABSsw ? H : H2;
	double hi, hv, db = Bmax/(double)max, dh = h2/(double)max;

	if (!ABSsw)
	    disp_center_line(g);
	g.setColor(Color.pink);
	g.drawString("0", xs, h2);

	for (int i = 1; i <= max; i++) {
	    hv = i*db;
            hi = i*dh;
            hm = (int)(h2-hi);
	    if (!ABSsw)
		hp = (int)(h2+hi);

	    g.setColor(Color.blue);
            g.drawLine(0, hm, W, hm);
	    if (!ABSsw)
		g.drawLine(0, hp, W, hp);

	    g.setColor(Color.pink);
            g.drawString(dform(hv), xs+2, hm+FNT);
	    if (!ABSsw)
		g.drawString("-"+dform(hv), xs, hp);
	}
    }
    public void disp_beat_vline(Graphics g, double dx, int dsf) {
        int wd;
        for (int i = 0; i < PA; i++) {
            wd = (int)(i*dx+dsf);

            g.setColor(Color.blue);
            g.drawLine(wd, 0, wd, H);

            g.setColor(Color.red);
            g.drawString(""+(i+1), wd, FNT*2);
        }
    }
    //
    public void disp_fmemori(Graphics g, int wmax) {
	int lmax = 10, hi, smax = Smplf/Wbyf;
	double ds = W/(double)smax;

	for(int i = 1; i < lmax; i++) { // h line
	    hi = HS*i/lmax; 
	    g.setColor(Color.blue);
	    g.drawLine(0, hi, W, hi);

	    g.setColor(Color.magenta);
	    g.drawString(""+(lmax-i), W-10, hi);
	}

	int me, di;
	for (int i = 0; i < smax; i += 10) { // disp memori
	    di = (int)(ds*i);
	    if (i%5000 == 0){
		if (smax >= 5000)
		    disp_k(g, i, di);
		me = 17;
	    } else if (i%1000 == 0){
		if (smax < 12000)
		    disp_k(g, i, di);
		me = 15;
	    } else if (i%500 == 0) {
		if (smax > 23000)
                    continue;
		me = 11;
	    } else if (i%100 == 0) {
		if (smax > 12000)
		    continue;
		me = 8;
	    } else if (smax < 2000){
		me = 5;
	    } else
		continue;
	    g.setColor(Color.gray);
	    g.drawLine(di, HS, di, HS+me);
	}
    }
    public void disp_k(Graphics g, int k, int x) {
        g.setColor(Color.orange);
        g.drawString(k/1000+"k", x, HS+FNT*2);
    }
    public void dispFft(Graphics g) {
	int x0 = 0, y0 = 0, x1, y1;
	int wmax = N/Wbyf;
	double dw = W/(double)wmax;

	disp_fmemori(g, wmax);
	calc_fft();

	g.setColor(Color.green);
	for (int i = 0; i < wmax; i++) {
	    x1 = (int)(dw*i);
	    y1 = (int)(Dat[i]*Gainf);
	    if (i == 0) {
		x0 = x1;
		y0 = y1;
	    }
	    g.drawLine(x0, HS-y0, x1, HS-y1);
	    x0 = x1;
	    y0 = y1;
	}
    }
    //------

    public double f2c(int keyn, double f) {
	return MC*Math.log(f/getFreq(keyn));
    }
    public double getFreq(int key) { // 0~87
        return AA*Math.pow(2, (key-48)/(double)Oct); // 49-1
    }
    public double geti(int key) {
	return MI.getInha(key);
    }
    public double getf(int key, int par) {
	return TT.get_ifreq(key, par);
    }
    public double getc(int key) {
	return TT.getCent(key);
    }
    public double getc(int key, int par) {
	return TT.get_icent(key, par); 
    }
    public double getpc(int key, int par) {
	return MI.getInha(key)*par*par;
    }
    public int ftok(double sf) {
        int key = (int)Math.rint(Oct*log_base(2, sf/getFreq(0)));
        if (key < 0)
            key = 0;
        else if (key >= KA)
            key = KA-1;
        return key;
    }
    public double log_base(double value) {
      return Math.log(value)/Log10;
    }
    public double log_base(double base, double value) {
        return Math.log(value)/Math.log(base);
    }
    public String getKeyn(int k) {
        return (k+1)+"("+getKey(k)+")";
    }
    public String getKey(int k) {
        return (k < 0) ? "" : Keynm[k%Oct];
    }
    public String getkey(boolean sw) {
	StringBuffer str = new StringBuffer(80);
	int key = sw ? KEY2 : KEY1;

	str.append(get_ch(sw)+":");
	str.append(getKeyn(key));
	str.append(dform(getf(key, 1))+"[Hz] ");
	str.append(dform(getc(key))+"[cent]");

	return str.toString();
    }
    //  
    public void re_write() {
	switch (GRAP) {
	case WAVE:
	    write_wave();
	    break;
	case BEAT:
	    write_beats();
	    break;
	case FFTX:
	    write_fft();
	    break;
	case CENT:
	    write_cent(CENT);
	    break;
	case INHA:
	    write_cent(INHA);
	    break;
	default:
	}

	text_mouse();
    }
    public void write_cent(int type) {
	StringBuffer str = new StringBuffer(1024);
	str.append("Key#:     [cent]");
	str.append("\n-------------------------\n");

	for (int i = 0; i < KB; i++) {
	    str.append(getKeyn(i));
	    str.append(df_form((type == CENT) ? getc(i) : geti(i)));
	    str.append("\n");
	}
	taA.setText(str.toString());
    }
    public void write_wave() {
	StringBuffer str = new StringBuffer(1024);
	String es = ": Key#: [cent]";
	String is = ": Inharmo.[cent]"+es;
	String ks = getkey(DSsw);
	int kn = get_key();

	str.append(ks).append("\nP#:      [Hz]");
	str.append(is);
	str.append("\n------------------------\n");
	double gf;
	int sck;
	for (int i = 1; i <= PA; i++) {
	    str.append(i+": ");
	    gf = getf(kn, i);
	    str.append(dform(gf));
	    str.append(df_form(getc(kn, i))+": ");
	    sck = ftok(gf);
	    str.append(getKeyn(sck));
	    str.append(df_form(f2c(sck, gf))+"\n");
	}
	taA.setText(str.toString());
    }
    public void write_beats() {
	StringBuffer str = new StringBuffer(1024);
	StringTokenizer tk, sp, bs;
	String line;
	double max;

	if (!CH2)
	    str.append(getkey(DSsw)+"\n");
	if (!CH1)
	    str.append(getkey(!DSsw)+"\n");
	str.append("P#:       [Hz]:   Beats\n");
	str.append("------------------------\n");
	for (int i = 1; i <= PA; i++) {
	    str.append(i);
	    str.append(df_form(DSsw ? Freq2[i] : Freq1[i]));
	    str.append(df_form(Beat[i]));
	    str.append("\n");
	}

	taA.setText(str.toString());
    }
    public void write_fft() {
	StringBuffer str = new StringBuffer(1024);
	str.append("Key#:     [Hz]");
	str.append(": Inharmo.[cent]");
	str.append("\n-------------------------\n");

	for (int i = 0; i < KB; i++) {
	    str.append(getKeyn(i));
	    str.append(df_form(getf(i, 1)));
	    str.append(df_form(getc(i))+"\n");
	}

	taA.setText(str.toString());
    }
    public String dform2(double val) {
	return df2.format((float)val);
    }
    public String dform(double val) {
	return df1.format((float)val);
    }
    public String df_form(double val) {
	return ": "+dform(val);
    }
    //-------------- fft -----------
    public void gauss() { // window
	double in;
	int max = (int)(Time*Smplf);

	for (int i = 0; i < N; i++) {
	    in = (2d*i-N)/N;
	    Im[i] = Re[i] = (i >= max) ? 0 :
		MW.getWave(i)*Math.exp(-4.5*in*in);
	}
    }
    public void fft() { // fft.java
        fftN.fft(Re, Im);
        for (int i = 0; i < N2; i++)
            Dat[i] = Math.sqrt(Re[i]*Re[i]+Im[i]*Im[i]);
    }
}
