/* $Id: MakeWaves2.java,v 1.3 2013/06/26 00:05:42 mit Exp $ */
/* MakeWaves2.java v0.1 for ifft2 */

public class MakeWaves2 {
    private final int KB = 88;
    private final int SP88 = 22; // 1/22
    private final int SP49 = 8;  // 1/8
    private final int K49 = 49-1;
    private final int SA =  KB-K49-1;
    private final double SAQ = Math.pow(SP88/(double)SP49, 1.0/SA);
    private final double CutoffL = 20.0;  // [Hz]
    private final double CutoffH = 170.0; //  "
    private final double PI2 = 2.0*Math.PI;

    private boolean SPsw;
    private boolean MAKEWAVEsw;
    private boolean CH_A = true;
    private boolean CH_B = true;
    private int Splgf;
    private int Multi;
    private int KeyA;
    private int KeyB;
    private int MaxData;   // SplgfxTime
    private double Time;
    private double TD;     // 1.0/Splgf;
    private double[] WaveX;

    public MakeWaves2() {}
    public MakeWaves2(int splf, double time, int mul) {
	initWave(splf, time, mul);
    }
    public void initWave(int splf, double time) {
	setTime(time);
	setSplgf(splf);
    }
    public void initWave(int splf, double time, int mul) {
	setTime(time);
	setSplgf(splf);
	setMulti(mul);
    }
    public void setSP(boolean sw) {
	SPsw = sw;

	makeWaveOff();
    }
    public void setCH(boolean a, boolean b) {
	CH_A = !a;
	CH_B = !b;

	makeWaveOff();
    }
    public void setSplgf(int splf) {
	Splgf = splf;
	TD = 1.0/Splgf;
	init();

	makeWaveOff();
    }
    public void setTime(double time) {
	Time = time;

	makeWaveOff();
    }
    public void setMulti(int mul) {
	Multi = mul;

	makeWaveOff();
    }
    public void setKey(int a, int b) {
	KeyA = a;
	KeyB = b;
 
	makeWaveOff();
    }
    public void init() {
	MaxData = (int)(Splgf*Time);
	initX();
    }
    public void initX() {
	WaveX = new double[MaxData];
    }
    //
    public void makeWaveOn() {
	MAKEWAVEsw = true;
    }
    public void makeWaveOff() {
	MAKEWAVEsw = false;
    }
    public boolean isMakeWave() {
	return MAKEWAVEsw;
    }
    public void makeWave(double[] freq) {
	double t;
	int i, j;

	for (j = 1; j <= Multi; j++) {
	    t = 0;
	    for (i = 0; i < MaxData; i++){
		WaveX[i] += Math.sin(freq[j]*t*PI2);
		t += TD;
	    }
	}
    }
    public void makeWaveSP(double[] freq, int key) {
	double t, ksp;
	int i, j;

	for (j = 1; j <= Multi; j++) {
	    ksp = calc_sindo(freq[j], j, key);
	    t = 0;
	    for (i = 0; i < MaxData; i++){
		WaveX[i] += Math.sin(freq[j]*t*PI2)*ksp;
		t += TD;
	    }
	}
    }
    public double getWave(int num) {
	return WaveX[num];
    }
    public void calcWave(double[] fa, double[] fb) {
	initX();

	if (SPsw) {
	    if (CH_A)
		makeWaveSP(fa, KeyA);
	    if(CH_B)
		makeWaveSP(fb, KeyB);
	} else {
	    if(CH_A)
		makeWave(fa);
	    if(CH_B)
		makeWave(fb);
	}

	makeWaveOn();
    }
    //
    public double calc_sindo(double freq, int pr, int key) {
        double fg = get_sboard(freq);
        double sp = get_spoint(key);
        double ga = spoint(pr, sp);

        return ga*fg;
    }
    public double get_spoint(int key) { // 0~481/8
        return (key <= K49) ?
	    (double)SP49 : SP49*Math.pow(SAQ, key-K49);
    }
    public double spoint(int mm, double sp) {
        double mp = mm * Math.PI;
        return Math.abs(8.0*Math.sin(mp/sp)/mp); // ̵· 8
    }
    public double get_sboard(double freq) {
        freq -= CutoffL;
        double gain = freq/CutoffH;

        if (gain < 0)
            gain = 0.0;
        else if (gain > 1.0)
            gain = 1.0;

        return gain;
    }
}